@@ -1,15 +1,31 @@
+"""
+A tide prediction toolkit for Python
+====================================
+
+pyTMD contains Python tools for reading OTIS, GOT and FES formatted tidal
+solutions to predict ocean and load tides
+
+The package works using scientific Python packages (numpy, scipy and pyproj)
+combined with data storage in netCDF4 and HDF5 and mapping using
+matplotlib and cartopy
+
+Documentation is available at https://pytmd.readthedocs.io
+"""
+import pyTMD.time
+import pyTMD.utilities
+from pyTMD.bilinear_interp import bilinear_interp
 from pyTMD.convert_julian import convert_julian
-from pyTMD.count_leap_seconds import count_leap_seconds
 from pyTMD.convert_calendar_decimal import convert_calendar_decimal
 from pyTMD.calc_delta_time import calc_delta_time
 from pyTMD.calc_astrol_longitudes import calc_astrol_longitudes
-from pyTMD.convert_xy_ll import convert_xy_ll
+from pyTMD.convert_ll_xy import convert_ll_xy
 from pyTMD.infer_minor_corrections import infer_minor_corrections
 from pyTMD.load_constituent import load_constituent
 from pyTMD.load_nodal_corrections import load_nodal_corrections
 from pyTMD.read_tide_model import extract_tidal_constants
 from pyTMD.read_netcdf_model import extract_netcdf_constants
 from pyTMD.read_GOT_model import extract_GOT_constants
+from pyTMD.read_FES_model import extract_FES_constants
 from pyTMD.predict_tidal_ts import predict_tidal_ts
 from pyTMD.predict_tide_drift import predict_tide_drift
 from pyTMD.predict_tide import predict_tide
